/*************************************************************
 * wave2d.c: solve  u_tt = u_xx + u_yy,  1999/08/14 tsakurai
 ------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <GL/glut.h>
#include "vector.h"  
#include "bmp.c"

/* default window size */
int width = 600, height = 400;

#define PI 3.14159265358979323846
#define DEG PI/180.0

#define IMAX 320
#define JMAX IMAX
#define NBUF 2
double  u[NBUF][IMAX+1][JMAX+1];

int NI = IMAX/2, NJ = JMAX/2;
Vector  vx[IMAX+1][JMAX+1];
Vector  vn[IMAX+1][JMAX+1];

/* num of drawing grid */
int num_grid = 40;

/* flags: drawing modes */
int smooth_shading = 1;
int animation = 1;
int frame = 1;
int wire  = 0;
int axis  = 1;
int clip  = 1;
int light = 3;
int dump_images = 0;
int dump_meshes = 0;

/* solving condition */
int initialize = 1; /* flag: 1 = do init, 0 = evolve */
int init_u = 0;
int bdry_condition = 1;
#define PERIODIC  0
#define DIRICHLET 1
#define NEUMANN   2 // not yet implemented

/* scaling */
GLfloat scale = 1.0, zmag = 1.;

/* eye = (r,theta,phi) in spherical coord */
double r = 7., theta = 70*DEG, phi   =-80*DEG;
double fovy = 12.;

/***************************************************
 * polygon utils
 --------------------------------------------------*/
void calcNormal(void)
{
    int i, j;
    for (j=0; j<=NJ; j++) for (i=0; i<=NI; i++) {
	static Vector O = {0, 0, 0}; vn[i][j] = O;
    }
    for (j=0; j<NJ; j++) for (i=0; i<NI; i++) {
	static Vector v0, v1, v2, n0, n1;
	VSub(&v0, &vx[i][j],     &vx[i][j+1]);
	VSub(&v1, &vx[i+1][j+1], &vx[i][j]);
	VSub(&v2, &vx[i+1][j],   &vx[i+1][j+1]);
	VCross(&n0, &v0, &v1);
	VAdd(&vn[i][j+1],  &vn[i][j+1],  &n0);
	VAdd(&vn[i][j],    &vn[i][j],    &n0);
	VAdd(&vn[i+1][j+1],&vn[i+1][j+1],  &n0);
	VCross(&n1, &v2, &v1);
	VAdd(&vn[i+1][j+1],&vn[i+1][j+1],&n1);
	VAdd(&vn[i][j],  &vn[i][j],  &n1);
	VAdd(&vn[i+1][j],  &vn[i+1][j],  &n1);
    }
    for (j=0; j<=NJ; j++)  for (i=0; i<=NI; i++) 
	VNormalize(&vn[i][j]);
}

void setVertex(double (*u)[IMAX+1][JMAX+1])
{
    double x0 = -1, x1 = 1;
    double y0 = -1, y1 = 1;
    int i, j;
    for (j=0; j<=NJ; j++) for (i=0; i<=NI; i++) {
	vx[i][j].x = x0 + (x1-x0)/NI*i;
	vx[i][j].y = y0 + (y1-y0)/NJ*j;
	vx[i][j].z = (*u)[i*IMAX/NI][j*JMAX/NJ]*zmag;
    }
    if (!wire) calcNormal();
}

void printVx(char *fname)
{
    int i, j;
    FILE *fp;

    fp = fopen(fname, "w");
    fprintf(fp, "# mesh\n");
    fprintf(fp, "%d * %d\n#\n", NI*3/5+1, NJ*3/5+1);
    for (j=NJ/5; j<=NJ*4/5; j++) for (i=NI/5; i<=NI*4/5; i++) {
	fprintf(fp, "%10.6f %10.6f %10.6f\n",
		vx[i][j].x,  vx[i][j].y, vx[i][j].z);
    }
    fclose(fp);
}

/***************************************************
 * Solving PDE
 --------------------------------------------------*/
void wave2d(void)
{
#define A  (44.)
#define L  (2.0)
#define dx (L/IMAX)
#define dy (L/JMAX)
#define dt (dx/sqrt(2))
#define X  (dx*i - L/2)  // -1 <= x <= 1
#define Y  (dy*j - L/2)  // -1 <= y <= 1
#define i(D)  ( (i+(D)+IMAX) % IMAX )
#define j(D)  ( (j+(D)+JMAX) % JMAX )
#define n(D)  ( (n+(D)+NBUF) % NBUF )
    
    static int n=0;
    int i, j;

    if (initialize) {
	initialize = 0;
	/* initial value */
	for (j=0; j<=JMAX; j++) {
	    for (i=0; i<=IMAX; i++) {
		if (init_u == 0) {
		    u[n( 0)][i][j] = u[n(-1)][i][j] =
			1./cosh(A*sqrt(X*X + Y*Y));
		} else if (init_u == 1) {
		    u[n(-1)][i][j] = -dt*
			20/cosh(A*sqrt(X*X + Y*Y));
		    u[n( 0)][i][j] = 0;
		} else { 
		    u[n( 0)][i][j] = u[n(-1)][i][j] = 
			.2*sin(PI/2*(X+1)*2)*sin(PI/2*(Y+1)*3);
		}
	    }
	}
	if (bdry_condition == DIRICHLET) {
	    for (j=0; j<=JMAX; j++) {
		u[n(-1)][0][j] = u[n(-1)][IMAX][j] = 0;
		u[n( 0)][0][j] = u[n( 0)][IMAX][j] = 0;
	    }
	    for (i=0; i<=IMAX; i++) {
		u[n(-1)][i][0] = u[n(-1)][i][JMAX] = 0;
		u[n( 0)][i][0] = u[n( 0)][i][JMAX] = 0;
	    }
	}
	setVertex(&u[n(0)]);
	glutPostRedisplay();
	printf("2dt = %f, 2dx = %f\n", 2*dt, 2*dx);
	return;
    }

    /* solving */
    for (j=0; j<=JMAX; j++) for (i=0; i<=IMAX; i++) {
	u[n(+1)][i][j]
      = 
	  - u[n(-1)][i][j]
          + (u[n][i(+1)][j] + u[n][i(-1)][j])/2
          + (u[n][i][j(+1)] + u[n][i][j(-1)])/2;
    }
    /* bdry condition */
    if (bdry_condition == DIRICHLET) {
	for (j=0; j<=JMAX; j++)
	    u[n(+1)][0][j] = u[n(+1)][IMAX][j] = 0;
	for (i=0; i<=IMAX; i++)
	    u[n(+1)][i][0] = u[n(+1)][i][JMAX] = 0;
    }

    /* draw */
    if ((n=(++n)%NBUF) == 0) {
	setVertex(&u[n(0)]);
	glutPostRedisplay();
	return;
    }
}

/***************************************************
 * display solution
 --------------------------------------------------*/
void drawSurface(void)
{
    int i, j;
    glEnable(GL_NORMALIZE);
    for (j=0; j<NJ; j++) {
	glBegin(GL_TRIANGLE_STRIP);
	for (i=0; i<=((j<NJ/2&&clip)? NI/2: NI); i++) {
	    glNormal3dv(&vn[i][j+1].x); glVertex3dv(&vx[i][j+1].x);
	    glNormal3dv(&vn[i][j].x)  ; glVertex3dv(&vx[i][j].x);
	}
	glEnd();
    }
}

void drawFrame(void)
{
    int i, j;
    for (j=0; j<=NJ; j+=NJ/num_grid) {
	glBegin(GL_LINE_STRIP);
	for (i=0; i<=((j<NJ/2&&clip)? NI/2: NI); i++)
	    glVertex3dv(&vx[i][j].x);
	glEnd();
    }
    for (i=0; i<=NI; i+=NI/num_grid) {
	glBegin(GL_LINE_STRIP);
	for (j=(i>NI/2&&clip)? NJ/2:0; j<=NJ; j++)
	    glVertex3dv(&vx[i][j].x);
	glEnd();
    }
}

void drawAxis(void)
{
    glDisable(GL_LIGHTING);
    glBegin(GL_LINES); {
	glColor3f(0.6, 0.0, 0.2);
	glVertex3f(-1.2, 0, 0);	glVertex3f(1.2, 0, 0);
	glColor3f(0.0, 0.4, 0.2);
	glVertex3f( 0,-1.2, 0); glVertex3f(0, 1.2, 0);
	// glColor3f(0.2, 0.0, 0.4);
	// glVertex3f( 0,  0, -1.2); glVertex3f(0, 0, 1.2);
    } glEnd();
    glEnable(GL_LIGHTING);
}

/***************************************************
 * glut call-back
 --------------------------------------------------*/
/* face Colors */
GLfloat material_color[]  = {.7, .6, .8, 1.0};
GLfloat specular_color[]  = {.4, .4, .4, 1.0};
GLfloat shininess = .4*128;
GLfloat mat_back_color[]  = {.6, .3, .4, 1.0};
GLfloat frame_color[]     = {.0, .0, .0, 1.0};
GLfloat wire_color[]      = {.5, .5, .7, 1.0};

/* Lighting */
GLfloat light0_ambient [] = {.1, .1, .1, 1.0};
GLfloat light0_diffuse [] = {.7, .7, .7, 1.0};
GLfloat light0_specular[] = {.5, .5, .5, 1.0};
GLfloat light0_position[] = {-15.0,-5.0, 20.0, 0.0};
GLfloat light1_diffuse [] = {.5, .5, .5, 1.0};
GLfloat light1_specular[] = {.5, .5, .5, 1.0};
GLfloat light1_position[] = {  5.0,-5.0, 0.0, 0.0};

void init(void)
{
    /* Z-Buffer */
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glEnable(GL_POLYGON_OFFSET_FILL);
    glPolygonOffset(1.0, 1.0);

    /* light properties */
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
	glLightfv(GL_LIGHT0, GL_AMBIENT,  light0_ambient);
	glLightfv(GL_LIGHT0, GL_DIFFUSE,  light0_diffuse);
	glLightfv(GL_LIGHT0, GL_SPECULAR, light0_specular);
    glEnable(GL_LIGHT1);
	glLightfv(GL_LIGHT1, GL_DIFFUSE,  light1_diffuse);
	glLightfv(GL_LIGHT1, GL_SPECULAR, light1_specular);

    /* surface properties */
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material_color);
    glMaterialfv(GL_FRONT, GL_SPECULAR, specular_color);
    glMaterialf (GL_FRONT, GL_SHININESS, shininess);
    glMaterialfv(GL_BACK , GL_AMBIENT_AND_DIFFUSE, mat_back_color);

    glEnable(GL_NORMALIZE);

    /* init surface */
    setVertex(&u[0]);
}

void draw(void)
{
    /* set camera */
    glMatrixMode(GL_PROJECTION); {
        glLoadIdentity();
        gluPerspective(fovy, (GLfloat)width/(GLfloat)height, 0.8*r, 1.2*r);
    }
    glMatrixMode(GL_MODELVIEW); {
        glLoadIdentity();
        gluLookAt(r*sin(theta)*cos(phi), r*sin(theta)*sin(phi), r*cos(theta),
		  0.0, 0.0, 0.0,
		  -cos(theta)*cos(phi), -cos(theta)*sin(phi), sin(theta));
    }

    /* put light */
    glLightfv(GL_LIGHT0, GL_POSITION, light0_position);
    glLightfv(GL_LIGHT1, GL_POSITION, light1_position);
    (light & 1) ? glEnable(GL_LIGHT0) : glDisable(GL_LIGHT0);
    (light & 2) ? glEnable(GL_LIGHT1) : glDisable(GL_LIGHT1);

    /* draw objects */
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glShadeModel((smooth_shading) ? GL_SMOOTH : (GL_FLAT));
    glPushMatrix(); {
	glScalef(scale, scale, scale);
	if (!wire) 
	    drawSurface();
	if (frame) {
	    glDisable(GL_LIGHTING);
	    glColor4fv((wire)? wire_color: frame_color);
	    drawFrame();
	    glEnable(GL_LIGHTING);
	}
    } glPopMatrix();

    if (axis) drawAxis();
    glutSwapBuffers();

    if (dump_images) { /* to make gif-animation */
	static int count = 0;
	char str[100];
	glReadBuffer(GL_FRONT);
	sprintf(str,"wave2d%02d.bmp", count);
	if (count % 3 == 0)
	    SaveImageAs(str);
	if (++count > 100) exit(0);
    }
    if (dump_meshes) {
	static int count = 0;
	char str[100];
	sprintf(str,"wave2d%02d.msh", count);
	if (   count == 0
	    || count == 6
	    || count ==15
	    || count ==27
	    || count ==42
	    ) 
	    printVx(str);
	if (++count > 80) exit(0);
    }
}

void idle(void) { if (animation) wave2d(); }
void reshape(int w, int h) { glViewport(0, 0, width=w, height=h); }

void key(unsigned char k, int x, int y)
{
    switch (k) {
    case'q': exit(0);
    case'1': light ^= 1; break;
    case'2': light ^= 2; break;
    case'a': axis = !axis; break;
    case'C': clip = !clip; break;
    case'c': animation = !animation; break;
    case'p': animation = !animation; break;
    case'f': frame = !frame; wire = 0; break;
    case'u': r *=.9; break;
    case'd': r /=.9; break;
    case'+': scale /=.9; break;
    case'-': scale *=.9; break;
    case';': scale *=.9; break;
    case'z': zmag /=.9; break;
    case'Z': zmag *=.9; break;
    case'r': initialize = 1; break;
    case'g': zmag = 1; initialize = 1; break;
    case's': smooth_shading = !smooth_shading; wire = 0; break;
    case'w': wire = !wire; frame = 1; return;
    case'*': if (NI/2>=num_grid) NJ = NI = NI/2; return;
    case'/': if (NI*2 <= IMAX)   NJ = NI = NI*2; return;
    case'i': init_u = (++init_u) % 3; initialize = 1; break;
    case'b': bdry_condition = (++bdry_condition) % 3; break;
    case'B': draw(); glReadBuffer(GL_FRONT); SaveImageAs("wave2d.bmp"); break;
    case'W': printVx("wave2d.msh"); break;
    default: return;
    }
    glutPostRedisplay();
}

void skey(int k, int x, int y)
{
    switch (k) {
    case GLUT_KEY_UP:    theta -= 2*DEG;  break;
    case GLUT_KEY_DOWN:	 theta += 2*DEG;  break;
    case GLUT_KEY_LEFT:	 phi   -= 2*DEG;  break;
    case GLUT_KEY_RIGHT: phi   += 2*DEG;  break;
    default: return;
    }
    glutPostRedisplay();
}

/***************************************************
 * Menu setup 
 --------------------------------------------------*/
void init_u_select(int value)
{ init_u = value; initialize = 1; glutPostRedisplay(); }

void bdry_select(int value)
{ bdry_condition = value; glutPostRedisplay(); }

void main_menu_select(int value)
{ key((unsigned char) value, 0, 0); glutPostRedisplay(); }

void make_menus(void)
{
    int init_menu, bdry_menu;
    init_menu = glutCreateMenu(init_u_select); {
	glutAddMenuEntry("u(0)=sech(r/e), u_t(0)=0", 0);
	glutAddMenuEntry("u(0)=0, u_t(0)=sech(r/e)", 1);
	glutAddMenuEntry("u(0)=sin(2x)sin(3y)",      2);
    }
    bdry_menu = glutCreateMenu(bdry_select); {
	glutAddMenuEntry("Dirichlet",  1);
	// glutAddMenuEntry("Neumann",    2);
	glutAddMenuEntry("Periodic",   0);
    }
    glutCreateMenu(main_menu_select); {
	glutAddSubMenu("Initial value",       init_menu);
	glutAddSubMenu("Bdry condition",      bdry_menu);
	glutAddMenuEntry("Restart",                 'r');
	glutAddMenuEntry("Animation (On/Off)",      'c');
	glutAddMenuEntry("Wireframe (On/Off)",      'w');
	glutAddMenuEntry("Draw Mesh (On/Off)",      'f');
	glutAddMenuEntry("1/4 Clipping(On/Off)",    'C');
	glutAddMenuEntry("Save Image (wave2d.bmp)", 'B');
	glutAddMenuEntry("Quit",                    'q');
    }
    glutAttachMenu(GLUT_RIGHT_BUTTON);
}

/***************************************************
 * main
 --------------------------------------------------*/
int main(int argc, char *argv[])
{
    int  idx;
    for (idx = 1; idx < argc; idx++) {
	if (strcmp(argv[idx], "-b") == 0) { dump_images = 1; continue; }
	if (strcmp(argv[idx], "-m") == 0) {
	    NI = NJ = IMAX; wire = 1; dump_meshes = 1; continue;
	}
	if (strcmp(argv[idx], "-g") == 0) {
	    if (sscanf(argv[++idx], "%dx%d", &width, &height) != 2)
		fprintf(stderr,"format: -g <width>x<height>\n"), exit(1);
	    continue;
	}
	/* else: show usage */
	fprintf(stderr, "\nwave2d [-g WxH] [-b] [-m]\n\n"
		"-b: dump each image to a bmp-file\n"
	        "-m: write vertex data to a mesh-file\n\n");
	exit(1);
    }

    /* initialize */
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
    glutInitWindowPosition(0, 0);
    glutInitWindowSize(width, height);
    glutCreateWindow(argv[0]);
    init();

    /* regist call-back functions */
    glutDisplayFunc(draw);
    glutReshapeFunc(reshape);
    glutIdleFunc(idle);
    glutKeyboardFunc(key);
    glutSpecialFunc(skey);
    make_menus();

    glutMainLoop();
    return 0;
}
